import java.io.*;
import java.util.*;

public class ImportPoiMap { 
    static final int MIN_STREET_TYPE = 0x2;
    static final int MAX_STREET_TYPE = 0x5;
    static final int METRO_TYPE = 0xb00;
    static final int WATER_TYPE = 0x46;
    static final double EPSILON = 0.000001;

    static final int SPEED[] = {80, 60, 40, 20}; // average speed km/h for streets at different levels

    public static void main(String[] args) throws Exception { 
        BufferedReader in = null;
        String line;
        TreeMap streets = new TreeMap();
        boolean translit = false;
        boolean smoothStreets = false;
        boolean pruneLevels = false;
        boolean crosses = false;
        boolean small = false;
        int type = 0;        
        String label = null;
        HashMap pointMap = new HashMap();

        for (int i = 0; i < args.length; i++) { 
            String opt = args[i];
            if (opt.startsWith("-")) { 
                if ("-crosses".equals(args[i])) { 
                    // generate data for calculation shortest path
                    crosses = true;
                } else if ("-small".equals(args[i])) { 
                    small = true;
                } else if ("-prune".equals(opt)) { 
                    pruneLevels = true;
                } else if ("-smooth".equals(opt)) { 
                    smoothStreets = true;
                } else if ("-translit".equals(opt)) { 
                    translit = true;
                }
            } else { 
                in = new BufferedReader(new FileReader(opt));
            }
        }

        while ((line = in.readLine()) != null) {
            if (line.startsWith("[END-RGN")) { 
                label = null;
            } else if (line.startsWith("Type=")) {
                type = Integer.parseInt(line.substring(7), 16);
            } else if (line.startsWith("Label=")) { 
                label = line.substring(6);
                if (translit) { 
                    label = Translit.translit2win(label);
                }
            } else if (label != null) { 
                if (type == METRO_TYPE) { 
                    if (line.startsWith("Origin1=")) { 
                        Street streetChain = (Street)streets.get(label);
                        Street street = streetChain;
                        while (street != null && street.type != ObjectType.METRO_STATION) {
                            street = street.collision;
                        }
                        if (street == null) { 
                            street = new Street(label, 0, ObjectType.METRO_STATION);
                            street.collision = streetChain;
                            streets.put(label, street);  
                        }
                        Segment segment = new Segment(street);
                        street.segments.add(segment);
                        int pos = 9;
                        int comma = line.indexOf(',', pos);
                        int lbr = line.indexOf(')', comma+1);
                        double x = Double.parseDouble(line.substring(comma+1, lbr));
                        double y = Double.parseDouble(line.substring(pos, comma));
                        segment.points.add(new Point(x, y, segment));
                    }
                } else if (type == WATER_TYPE && !small) { 
                    if (line.startsWith("Data")) {
                        char dataType = line.charAt(4); 
                        Street streetChain = (Street)streets.get(label);
                        Street street = streetChain;
                        while (street != null && street.type != ObjectType.WATER) {
                            street = street.collision;
                        }
                        int streetLevel = 0;
                        if (street == null) {
                            street = new Street(label, streetLevel, ObjectType.WATER);
                            street.dataType = dataType;
                            streets.put(label, street);  
                            street.collision = streetChain;
                        } else if (street.dataType > dataType) { 
                            if (pruneLevels) { 
                                // For Moscow to reduce jar size we have to use the least detailed level
                                continue;
                            } else { 
                                // And for St.Peterburg it is possible to use most detailed level
                                street.dataType = dataType;
                                street.segments.clear();         
                            }
                        } else if (street.dataType != dataType) { 
                            throw new Error("Unexpected data type");
                        }                        

                        Segment segment = new Segment(street);
                        street.segments.add(segment);
                        for (int pos = 7, len = line.length(); pos < len; ) { 
                            int comma = line.indexOf(',', pos);
                            int lbr = line.indexOf(')', comma+1);
                            double x = Double.parseDouble(line.substring(comma+1, lbr));
                            double y = Double.parseDouble(line.substring(pos, comma));
                            Point p = new Point(x, y, segment);
                            segment.points.add(p);
                            pos = lbr + 3;
                        }
                        segment.points.add(segment.points.get(0));
                    }
                } else if (type >= MIN_STREET_TYPE && type <= MAX_STREET_TYPE) { 
                    if (line.startsWith("Data")) {
                        char dataType = line.charAt(4); 
                        Street streetChain = (Street)streets.get(label);
                        Street street = streetChain;
                        int streetLevel = type - MIN_STREET_TYPE;
                        while (street != null && (street.type != ObjectType.STREET || street.level != streetLevel)) {
                            street = street.collision;
                        }
                        if (street == null) { 
                            street = new Street(label, streetLevel, ObjectType.STREET);
                            street.dataType = dataType;
                            street.collision = streetChain;
                            streets.put(label, street);  
                        } else if (street.dataType > dataType) { 
                            if (pruneLevels) { 
                                // For Moscow to reduce jar size we have to use the least detailed level
                                continue;
                            } else { 
                                // And for St.Peterburg it is possible to use most detailed level
                                street.dataType = dataType;
                                street.segments.clear();         
                            }
                        } else if (street.dataType != dataType) { 
                            throw new Error("Unexpected data type");
                        }
                        Segment segment = new Segment(street);
                        street.segments.add(segment);
                        for (int pos = 7, len = line.length(); pos < len; ) { 
                            int comma = line.indexOf(',', pos);
                            int lbr = line.indexOf(')', comma+1);
                            double x = Double.parseDouble(line.substring(comma+1, lbr));
                            double y = Double.parseDouble(line.substring(pos, comma));
                            Point p = new Point(x, y, segment);
                            segment.points.add(p);
                            Counter counter = (Counter)pointMap.get(p);
                            if (counter == null) { 
                                counter = new Counter();
                                pointMap.put(p, counter);
                            }
                            counter.value += 1;
                            pos = lbr + 3;
                        }
                    }
                }
            }
        }

        in.close();
        int nStreets = 0;
        Iterator iterator = streets.values().iterator();
        while (iterator.hasNext()) { 
            Street chain = (Street)iterator.next();
            for (Street s = chain; s != null; s = s.collision) { 
                nStreets += 1;
                if (s.type != ObjectType.STREET) { 
                    continue;
                }
                ArrayList segments = s.segments;
              merge:
                while (segments.size() > 1) {
                    int n = segments.size();
                    for (int i = 0; i < n; i++) { 
                        Segment s0 = (Segment)segments.get(i);
                        Point p0 = (Point)s0.points.get(s0.points.size()-1);
                        for (int j = 0; j < n; j++) { 
                            if (i != j) { 
                                Segment s1 = (Segment)segments.get(j);
                                Point p1 = (Point)s1.points.get(0);
                                if (p0.x == p1.x && p0.y == p1.y) { 
                                    for (int k = 1; k < s1.points.size(); k++) { 
                                        s0.points.add(s1.points.get(k));
                                    }
                                    Counter counter = (Counter)pointMap.get(p0);
                                    if (counter != null) { 
                                        counter.value -= 1;
                                    }
                                    segments.remove(j);
                                    continue merge;
                                }
                            }
                        }                    
                    }
                    break;
                }
                if (smoothStreets) { 
                    for (int i = 0; i < segments.size(); i++) { 
                        ArrayList points = ((Segment)segments.get(i)).points;
                        int n = points.size();                        
                        if (n > 3) { 
                            Point p0 = (Point)points.get(1);                    
                            n -= 1;
                            for (int k = 2; k < n; k++) {
                                Point p1 = (Point)points.get(k);
                                if (!crosses || ((Counter)pointMap.get(p0)).value == 1) { 
                                    p0.x = (p1.x + p0.x) / 2.0;
                                    p0.y = (p1.y + p0.y) / 2.0;
                                }
                                p0 = p1;
                            }
                            if (!crosses || ((Counter)pointMap.get(p0)).value == 1) {
                                points.remove(n-1);
                            } else { 
                                n += 1;
                            }
                        }
                        if (n > 2) { 
                            Point p0 = (Point)points.get(0);                    
                            Point p1 = (Point)points.get(1);                    
                            for (int k = 2; k < n;) {
                                Point p2 = (Point)points.get(k);
                                double a = (p1.x-p0.x)*(p1.x-p0.x) + (p1.y-p0.y)*(p1.y-p0.y);
                                double b = (p2.x-p1.x)*(p2.x-p1.x) + (p2.y-p1.y)*(p2.y-p1.y);
                                double c = (p2.x-p0.x)*(p2.x-p0.x) + (p2.y-p0.y)*(p2.y-p0.y);
                                double d = 2*(a*c + a*b + b*c) - a*a - b*b - c*c;
                                Counter counter = (Counter)pointMap.get(p1);
                                //if (d/c < EPSILON && (!crosses || (counter == null || counter.value == 1))) { 
                                if ((p1.x-p0.x)*(p2.y-p1.y) == (p1.y-p0.y)*(p2.x-p1.x) 
                                    && (!crosses || (counter == null || counter.value == 1))) 
                                { 
                                    points.remove(k-1);
                                    p1 = p2;
                                    n -= 1;
                                } else { 
                                    p0 = p1;
                                    p1 = p2;
                                    k += 1;
                                }
                            }
                            if (points.size() < 2) throw new Error("ERROR");
                        }
                    }            
                }
            }
        }
        MapWriter.write(nStreets, streets, new LatitudeLongitudeGrid(), SPEED, crosses, small);
    }
}